# Generated by Django 2.2.24 on 2021-08-12 10:44

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models

import mozilla_django_oidc_db.models


class Migration(migrations.Migration):

    dependencies = [
        ("mozilla_django_oidc_db", "0003_auto_20210719_0803"),
    ]

    operations = [
        migrations.AddField(
            model_name="openidconnectconfig",
            name="claim_mapping",
            field=django.contrib.postgres.fields.jsonb.JSONField(
                default=mozilla_django_oidc_db.models.get_claim_mapping,
                help_text="Mapping from user-model fields to OIDC claims",
                verbose_name="claim mapping",
            ),
        ),
        migrations.AddField(
            model_name="openidconnectconfig",
            name="groups_claim",
            field=models.CharField(
                default="roles",
                help_text="The name of the OIDC claim that holds the values to map to local user groups.",
                max_length=50,
                verbose_name="groups claim",
            ),
        ),
        migrations.AddField(
            model_name="openidconnectconfig",
            name="make_users_staff",
            field=models.BooleanField(
                default=False,
                help_text="Users will be flagged as being a staff user automatically. This allows users to login to the admin interface. By default they have no permissions, even if they are staff.",
                verbose_name="make users staff",
            ),
        ),
        migrations.AddField(
            model_name="openidconnectconfig",
            name="sync_groups",
            field=models.BooleanField(
                default=True,
                help_text="Synchronize the local user groups with the provided groups. Note that this means a user is removed from all groups if there is no group claim. Uncheck to manage groups manually.",
                verbose_name="synchronize groups",
            ),
        ),
    ]



# Generated by Django 2.2.24 on 2021-07-05 10:16

from django.db import migrations, models

import django_better_admin_arrayfield.models.fields

import mozilla_django_oidc_db.models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="OpenIDConnectConfig",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "enabled",
                    models.BooleanField(
                        default=False,
                        help_text="Indicates whether OpenID Connect for authentication/authorization is enabled",
                        verbose_name="enable",
                    ),
                ),
                (
                    "oidc_rp_client_id",
                    models.CharField(
                        help_text="OpenID Connect client ID provided by the OIDC Provider",
                        max_length=1000,
                        verbose_name="OpenID Connect client ID",
                    ),
                ),
                (
                    "oidc_rp_client_secret",
                    models.CharField(
                        help_text="OpenID Connect secret provided by the OIDC Provider",
                        max_length=1000,
                        verbose_name="OpenID Connect secret",
                    ),
                ),
                (
                    "oidc_rp_sign_algo",
                    models.CharField(
                        default="HS256",
                        help_text="Algorithm the Identity Provider uses to sign ID tokens",
                        max_length=50,
                        verbose_name="OpenID sign algorithm",
                    ),
                ),
                (
                    "oidc_rp_scopes_list",
                    django_better_admin_arrayfield.models.fields.ArrayField(
                        base_field=models.CharField(
                            max_length=50, verbose_name="OpenID Connect scope"
                        ),
                        blank=True,
                        default=mozilla_django_oidc_db.models.get_default_scopes,
                        help_text="OpenID Connect scopes that are requested during login",
                        size=None,
                        verbose_name="OpenID Connect scopes",
                    ),
                ),
                (
                    "oidc_op_jwks_endpoint",
                    models.URLField(
                        blank=True,
                        help_text="URL of your OpenID Connect provider JSON Web Key Set endpoint. Required if `RS256` is used as signing algorithm",
                        max_length=1000,
                        verbose_name="JSON Web Key Set endpoint",
                    ),
                ),
                (
                    "oidc_op_authorization_endpoint",
                    models.URLField(
                        help_text="URL of your OpenID Connect provider authorization endpoint",
                        max_length=1000,
                        verbose_name="Authorization endpoint",
                    ),
                ),
                (
                    "oidc_op_token_endpoint",
                    models.URLField(
                        help_text="URL of your OpenID Connect provider token endpoint",
                        max_length=1000,
                        verbose_name="Token endpoint",
                    ),
                ),
                (
                    "oidc_op_user_endpoint",
                    models.URLField(
                        help_text="URL of your OpenID Connect provider userinfo endpoint",
                        max_length=1000,
                        verbose_name="User endpoint",
                    ),
                ),
                (
                    "oidc_rp_idp_sign_key",
                    models.CharField(
                        blank=True,
                        help_text="Key the Identity Provider uses to sign ID tokens in the case of an RSA sign algorithm. Should be the signing key in PEM or DER format",
                        max_length=1000,
                        verbose_name="Sign key",
                    ),
                ),
            ],
            options={
                "verbose_name": "OpenID Connect configuration",
            },
        ),
    ]



import os

BASE_DIR = os.path.abspath(os.path.dirname(__file__))

SECRET_KEY = "so-secret-i-cant-believe-you-are-looking-at-this"

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.postgresql",
        "NAME": os.getenv("PGDATABASE", "mozilla_django_oidc_db"),
        "USER": os.getenv("PGUSER", "mozilla_django_oidc_db"),
        "PASSWORD": os.getenv("PGPASSWORD", "mozilla_django_oidc_db"),
        "HOST": os.getenv("DB_HOST", "localhost"),
        "PORT": os.getenv("DB_PORT", 5432),
    }
}

CACHES = {
    "default": {"BACKEND": "django.core.cache.backends.locmem.LocMemCache"},
    "oidc": {"BACKEND": "django.core.cache.backends.locmem.LocMemCache"},
}

INSTALLED_APPS = [
    "django.contrib.contenttypes",
    "django.contrib.auth",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.admin",
    "mozilla_django_oidc",
    "mozilla_django_oidc_db",
    "testapp",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

ROOT_URLCONF = "testapp.urls"



from django.test import RequestFactory

import pytest

from mozilla_django_oidc_db.middleware import SessionRefresh
from mozilla_django_oidc_db.models import OpenIDConnectConfig


@pytest.mark.django_db
def test_sessionrefresh_oidc_not_enabled():
    config = OpenIDConnectConfig.get_solo()

    config.enabled = False
    config.save()

    request = RequestFactory().get("/")

    # Running the middleware should return None, since OIDC is disabled
    result = SessionRefresh(get_response=lambda: None).process_request(request)

    assert result is None



from json.decoder import JSONDecodeError
from unittest.mock import patch

from django.utils.translation import gettext as _

import pytest
import requests_mock
from requests import Response
from requests.exceptions import RequestException

from mozilla_django_oidc_db.forms import OpenIDConnectConfigForm
from mozilla_django_oidc_db.models import OpenIDConnectConfig, get_claim_mapping


@pytest.mark.django_db
def test_derive_endpoints_success():
    form_data = {
        "oidc_rp_client_id": "clientid",
        "oidc_rp_client_secret": "secret",
        "oidc_rp_sign_algo": "RS256",
        "oidc_op_discovery_endpoint": "http://discovery-endpoint.nl/",
        "claim_mapping": get_claim_mapping(),
        "groups_claim": "roles",
        "sync_groups_glob_pattern": "*",
        "username_claim": "sub",
    }
    form = OpenIDConnectConfigForm(data=form_data)

    configuration = {
        "authorization_endpoint": "http://provider.com/auth/realms/master/protocol/openid-connect/auth",
        "token_endpoint": "http://provider.com/auth/realms/master/protocol/openid-connect/token",
        "userinfo_endpoint": "http://provider.com/auth/realms/master/protocol/openid-connect/userinfo",
        "jwks_uri": "http://provider.com/auth/realms/master/protocol/openid-connect/certs",
    }
    with requests_mock.Mocker() as m:
        m.get(
            "http://discovery-endpoint.nl/.well-known/openid-configuration",
            json=configuration,
        )
        assert form.is_valid()

    config = form.save()
    assert (
        config.oidc_op_authorization_endpoint
        == "http://provider.com/auth/realms/master/protocol/openid-connect/auth"
    )
    assert (
        config.oidc_op_token_endpoint
        == "http://provider.com/auth/realms/master/protocol/openid-connect/token"
    )
    assert (
        config.oidc_op_user_endpoint
        == "http://provider.com/auth/realms/master/protocol/openid-connect/userinfo"
    )
    assert (
        config.oidc_op_jwks_endpoint
        == "http://provider.com/auth/realms/master/protocol/openid-connect/certs"
    )


@pytest.mark.django_db
def test_derive_endpoints_extra_field():
    form_data = {
        "oidc_rp_client_id": "clientid",
        "oidc_rp_client_secret": "secret",
        "oidc_rp_sign_algo": "RS256",
        "oidc_op_discovery_endpoint": "http://discovery-endpoint.nl/",
        "claim_mapping": get_claim_mapping(),
        "groups_claim": "roles",
        "sync_groups_glob_pattern": "*",
        "username_claim": "sub",
    }

    class ExtendedOpenIDConnectConfigForm(OpenIDConnectConfigForm):
        required_endpoints = OpenIDConnectConfigForm.required_endpoints
        # Define an extra field to derive from the configuration
        oidc_mapping = dict(
            **OpenIDConnectConfigForm.oidc_mapping,
            **{"logout_endpoint": "end_session_endpoint"}
        )

    form = ExtendedOpenIDConnectConfigForm(data=form_data)

    configuration = {
        "authorization_endpoint": "http://provider.com/auth/realms/master/protocol/openid-connect/auth",
        "token_endpoint": "http://provider.com/auth/realms/master/protocol/openid-connect/token",
        "userinfo_endpoint": "http://provider.com/auth/realms/master/protocol/openid-connect/userinfo",
        "jwks_uri": "http://provider.com/auth/realms/master/protocol/openid-connect/certs",
        "end_session_endpoint": "http://provider.com/auth/realms/master/protocol/openid-connect/logout",
    }
    with requests_mock.Mocker() as m:
        m.get(
            "http://discovery-endpoint.nl/.well-known/openid-configuration",
            json=configuration,
        )
        assert form.is_valid()

    # The endpoint that was added to the mapping on the extended form
    # should be present in the cleaned data
    assert (
        form.cleaned_data["logout_endpoint"]
        == "http://provider.com/auth/realms/master/protocol/openid-connect/logout"
    )


@patch("requests.get", side_effect=RequestException)
def test_derive_endpoints_request_error(*m):
    form_data = {
        "oidc_rp_client_id": "clientid",
        "oidc_rp_client_secret": "secret",
        "oidc_rp_sign_algo": "RS256",
        "oidc_op_discovery_endpoint": "http://discovery-endpoint.nl",
        "claim_mapping": get_claim_mapping(),
        "groups_claim": "roles",
        "sync_groups_glob_pattern": "*",
        "username_claim": "sub",
    }
    form = OpenIDConnectConfigForm(data=form_data)

    form.is_valid()

    assert form.errors == {
        "oidc_op_discovery_endpoint": [
            _("Something went wrong while retrieving the configuration.")
        ]
    }


@patch("requests.get", side_effect=JSONDecodeError("error", "test", 1))
def test_derive_endpoints_json_error(*m):
    form_data = {
        "oidc_rp_client_id": "clientid",
        "oidc_rp_client_secret": "secret",
        "oidc_rp_sign_algo": "RS256",
        "oidc_op_discovery_endpoint": "http://discovery-endpoint.nl",
        "claim_mapping": get_claim_mapping(),
        "groups_claim": "roles",
        "sync_groups_glob_pattern": "*",
        "username_claim": "sub",
    }
    form = OpenIDConnectConfigForm(data=form_data)

    form.is_valid()

    assert form.errors == {
        "oidc_op_discovery_endpoint": [
            _("Something went wrong while retrieving the configuration.")
        ]
    }


def test_no_discovery_endpoint_other_fields_required():
    form_data = {
        "oidc_rp_client_id": "clientid",
        "oidc_rp_client_secret": "secret",
        "oidc_rp_sign_algo": "RS256",
        "claim_mapping": get_claim_mapping(),
        "groups_claim": "roles",
        "sync_groups_glob_pattern": "*",
        "username_claim": "sub",
    }
    form = OpenIDConnectConfigForm(data=form_data)

    form.is_valid()

    assert form.errors == {
        "oidc_op_authorization_endpoint": [_("This field is required.")],
        "oidc_op_token_endpoint": [_("This field is required.")],
        "oidc_op_user_endpoint": [_("This field is required.")],
    }


def test_admin_form_readonly_access():
    # Empty the base_fields, causing OpenIDConnectConfigForm.fields to be empty
    # as well, which is also the case for when users access the form with
    # read only access
    OpenIDConnectConfigForm.base_fields = {}

    # Form initialization should not raise any errors
    form = OpenIDConnectConfigForm()



import pytest

from mozilla_django_oidc_db.backends import OIDCAuthenticationBackend
from mozilla_django_oidc_db.middleware import SessionRefresh
from mozilla_django_oidc_db.models import OpenIDConnectConfig
from mozilla_django_oidc_db.views import OIDCAuthenticationRequestView


@pytest.mark.django_db
def test_backend_settings_derived_from_model_oidc_not_enabled():
    config = OpenIDConnectConfig.get_solo()
    config.enabled = False
    config.save()

    backend = OIDCAuthenticationBackend()

    # verify that the settings are not set (because of early return in __init__)
    assert not hasattr(backend, "OIDC_RP_CLIENT_ID")
    assert not hasattr(backend, "OIDC_RP_CLIENT_SECRET")
    assert not hasattr(backend, "OIDC_RP_SIGN_ALGO")
    assert not hasattr(backend, "OIDC_OP_JWKS_ENDPOINT")
    assert not hasattr(backend, "OIDC_OP_TOKEN_ENDPOINT")
    assert not hasattr(backend, "OIDC_OP_USER_ENDPOINT")
    assert not hasattr(backend, "OIDC_RP_IDP_SIGN_KEY")


@pytest.mark.django_db
def test_backend_settings_derived_from_model_oidc_enabled():
    config = OpenIDConnectConfig.get_solo()

    config.enabled = True
    config.oidc_rp_client_id = "testid"
    config.oidc_rp_client_secret = "secret"
    config.oidc_rp_sign_algo = "HS256"
    config.oidc_rp_scopes_list = ["openid", "email"]
    config.oidc_op_jwks_endpoint = "http://some.endpoint/v1/jwks"
    config.oidc_op_authorization_endpoint = "http://some.endpoint/v1/auth"
    config.oidc_op_token_endpoint = "http://some.endpoint/v1/token"
    config.oidc_op_user_endpoint = "http://some.endpoint/v1/user"

    config.save()

    backend = OIDCAuthenticationBackend()

    # verify that the settings are derived from OpenIDConnectConfig
    assert backend.OIDC_RP_CLIENT_ID == "testid"
    assert backend.OIDC_RP_CLIENT_SECRET == "secret"
    assert backend.OIDC_RP_SIGN_ALGO == "HS256"
    assert backend.OIDC_OP_JWKS_ENDPOINT == "http://some.endpoint/v1/jwks"
    assert backend.OIDC_OP_TOKEN_ENDPOINT == "http://some.endpoint/v1/token"
    assert backend.OIDC_OP_USER_ENDPOINT == "http://some.endpoint/v1/user"
    assert backend.OIDC_RP_IDP_SIGN_KEY is None


@pytest.mark.django_db
def test_view_settings_derived_from_model_oidc_enabled():
    config = OpenIDConnectConfig.get_solo()

    config.enabled = True
    config.oidc_rp_client_id = "testid"
    config.oidc_rp_client_secret = "secret"
    config.oidc_rp_sign_algo = "HS256"
    config.oidc_rp_scopes_list = ["openid", "email"]
    config.oidc_op_jwks_endpoint = "http://some.endpoint/v1/jwks"
    config.oidc_op_authorization_endpoint = "http://some.endpoint/v1/auth"
    config.oidc_op_token_endpoint = "http://some.endpoint/v1/token"
    config.oidc_op_user_endpoint = "http://some.endpoint/v1/user"

    config.save()

    view = OIDCAuthenticationRequestView()

    # verify that the settings are derived from OpenIDConnectConfig
    assert view.OIDC_RP_CLIENT_ID == "testid"
    assert view.OIDC_OP_AUTH_ENDPOINT == "http://some.endpoint/v1/auth"



from mozilla_django_oidc.middleware import SessionRefresh
from mozilla_django_oidc.utils import import_from_settings

from .mixins import SoloConfigMixin
from .models import OpenIDConnectConfig


class SessionRefresh(SoloConfigMixin, SessionRefresh):
    def process_request(self, request):
        config = OpenIDConnectConfig.get_solo()
        if not config.enabled:
            return

        return super().process_request(request)



# The cache that should be used, e.g. 'default'. Refers to Django CACHES setting.
# Set to None to disable caching.
MOZILLA_DJANGO_OIDC_DB_CACHE = "oidc"

MOZILLA_DJANGO_OIDC_DB_CACHE_TIMEOUT = 60 * 5

MOZILLA_DJANGO_OIDC_DB_PREFIX = "oidc"



import fnmatch
import logging

from django.contrib.auth.models import Group
from django.core.exceptions import ObjectDoesNotExist

from mozilla_django_oidc.auth import (
    OIDCAuthenticationBackend as _OIDCAuthenticationBackend,
)

from .mixins import SoloConfigMixin
from .models import OpenIDConnectConfig

logger = logging.getLogger(__name__)


class OIDCAuthenticationBackend(SoloConfigMixin, _OIDCAuthenticationBackend):
    """
    Modifies the default OIDCAuthenticationBackend to use a configurable claim
    as unique identifier (default `sub`).
    """

    def __init__(self, *args, **kwargs):
        self.config = OpenIDConnectConfig.get_solo()

        if not self.config.enabled:
            return

        super().__init__(*args, **kwargs)

    def authenticate(self, *args, **kwargs):
        if not self.config.enabled:
            return None

        return super().authenticate(*args, **kwargs)

    def get_user_instance_values(self, claims):
        """
        Map the names and values of the claims to the fields of the User model
        """
        return {
            model_field: claims.get(claims_field, "")
            for model_field, claims_field in self.config.claim_mapping.items()
        }

    def create_user(self, claims):
        """Return object for a newly created user account."""
        username_claim = self.config.username_claim
        unique_id = claims.get(username_claim)

        logger.debug("Creating OIDC user: %s", unique_id)

        user = self.UserModel.objects.create_user(
            **{self.UserModel.USERNAME_FIELD: unique_id}
        )
        self.update_user(user, claims)

        return user

    def filter_users_by_claims(self, claims):
        """Return all users matching the specified subject."""
        username_claim = self.config.username_claim
        unique_id = claims.get(username_claim)

        if not unique_id:
            return self.UserModel.objects.none()
        return self.UserModel.objects.filter(
            **{f"{self.UserModel.USERNAME_FIELD}__iexact": unique_id}
        )

    def verify_claims(self, claims):
        """Verify the provided claims to decide if authentication should be allowed."""
        scopes = self.get_settings("OIDC_RP_SCOPES", "openid email")

        logger.debug("OIDC claims received: %s", claims)

        username_claim = self.config.username_claim

        if username_claim not in claims:
            logger.error(
                "%s not in OIDC claims, cannot proceed with authentication",
                username_claim,
            )
            return False
        return True

    def update_user(self, user, claims):
        """Update existing user with new claims, if necessary save, and return user"""
        values = self.get_user_instance_values(claims)
        for field, value in values.items():
            setattr(user, field, value)
        logger.debug("Updating OIDC user %s with: %s", user, values)

        # Users can only be promoted to staff. Staff rights are never taken by OIDC.
        if self.config.make_users_staff and not user.is_staff:
            user.is_staff = True
            user.save(update_fields=["is_staff"])

        user.save(update_fields=values.keys())

        self.update_user_groups(user, claims)

        return user

    def update_user_groups(self, user, claims):
        """
        Updates user group memberships based on the group_claim setting.

        Copied and modified from: https://github.com/snok/django-auth-adfs/blob/master/django_auth_adfs/backend.py
        """
        groups_claim = self.config.groups_claim

        if groups_claim:
            # Update the user's group memberships
            django_groups = [group.name for group in user.groups.all()]

            if groups_claim in claims:
                claim_groups = claims[groups_claim]
                if not isinstance(claim_groups, list):
                    claim_groups = [
                        claim_groups,
                    ]
            else:
                logger.debug(
                    "The configured groups claim '%s' was not found in the access token",
                    groups_claim,
                )
                claim_groups = []
            if sorted(claim_groups) != sorted(django_groups):
                existing_groups = list(
                    Group.objects.filter(name__in=claim_groups).iterator()
                )
                existing_group_names = frozenset(
                    group.name for group in existing_groups
                )
                new_groups = []
                if self.config.sync_groups:
                    # Only sync groups that match the supplied glob pattern
                    new_groups = [
                        Group.objects.get_or_create(name=name)[0]
                        for name in fnmatch.filter(
                            claim_groups, self.config.sync_groups_glob_pattern
                        )
                        if name not in existing_group_names
                    ]
                else:
                    for name in claim_groups:
                        if name not in existing_group_names:
                            try:
                                group = Group.objects.get(name=name)
                                new_groups.append(group)
                            except ObjectDoesNotExist:
                                pass
                user.groups.set(existing_groups + new_groups)



from typing import Dict

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.postgres.fields import JSONField
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.encoding import force_text
from django.utils.translation import gettext_lazy as _

from django_better_admin_arrayfield.models.fields import ArrayField
from solo.models import SingletonModel, get_cache

import mozilla_django_oidc_db.settings as oidc_settings


def get_default_scopes():
    """
    Returns the default scopes to request for OpenID Connect logins
    """
    return ["openid", "email", "profile"]


def get_claim_mapping() -> Dict[str, str]:
    # Map (some) claim names from https://openid.net/specs/openid-connect-core-1_0.html#Claims
    # to corresponding field names on the User model
    return {
        "email": "email",
        "first_name": "given_name",
        "last_name": "family_name",
    }


class OpenIDConnectConfig(SingletonModel):
    """
    Configuration for authentication/authorization via OpenID connect
    """

    enabled = models.BooleanField(
        _("enable"),
        default=False,
        help_text=_(
            "Indicates whether OpenID Connect for authentication/authorization is enabled"
        ),
    )

    oidc_rp_client_id = models.CharField(
        _("OpenID Connect client ID"),
        max_length=1000,
        help_text=_("OpenID Connect client ID provided by the OIDC Provider"),
    )
    oidc_rp_client_secret = models.CharField(
        _("OpenID Connect secret"),
        max_length=1000,
        help_text=_("OpenID Connect secret provided by the OIDC Provider"),
    )
    oidc_rp_sign_algo = models.CharField(
        _("OpenID sign algorithm"),
        max_length=50,
        help_text=_("Algorithm the Identity Provider uses to sign ID tokens"),
        default="HS256",
    )
    oidc_rp_scopes_list = ArrayField(
        verbose_name=_("OpenID Connect scopes"),
        base_field=models.CharField(_("OpenID Connect scope"), max_length=50),
        default=get_default_scopes,
        blank=True,
        help_text=_("OpenID Connect scopes that are requested during login"),
    )

    oidc_op_discovery_endpoint = models.URLField(
        _("Discovery endpoint"),
        max_length=1000,
        help_text=_(
            "URL of your OpenID Connect provider discovery endpoint ending with a slash "
            "(`.well-known/...` will be added automatically). "
            "If this is provided, the remaining endpoints can be omitted, as "
            "they will be derived from this endpoint."
        ),
        blank=True,
    )
    oidc_op_jwks_endpoint = models.URLField(
        _("JSON Web Key Set endpoint"),
        max_length=1000,
        help_text=_(
            "URL of your OpenID Connect provider JSON Web Key Set endpoint. Required if `RS256` is used as signing algorithm"
        ),
        blank=True,
    )
    oidc_op_authorization_endpoint = models.URLField(
        _("Authorization endpoint"),
        max_length=1000,
        help_text=_("URL of your OpenID Connect provider authorization endpoint"),
    )
    oidc_op_token_endpoint = models.URLField(
        _("Token endpoint"),
        max_length=1000,
        help_text=_("URL of your OpenID Connect provider token endpoint"),
    )
    oidc_op_user_endpoint = models.URLField(
        _("User endpoint"),
        max_length=1000,
        help_text=_("URL of your OpenID Connect provider userinfo endpoint"),
    )
    oidc_rp_idp_sign_key = models.CharField(
        _("Sign key"),
        max_length=1000,
        help_text=_(
            "Key the Identity Provider uses to sign ID tokens in the case of an RSA sign algorithm. Should be the signing key in PEM or DER format"
        ),
        blank=True,
    )

    username_claim = models.CharField(
        _("username claim"),
        max_length=50,
        default="sub",
        help_text=_("The name of the OIDC claim that is used as the username"),
    )
    claim_mapping = JSONField(
        _("claim mapping"),
        default=get_claim_mapping,
        help_text=("Mapping from user-model fields to OIDC claims"),
    )
    groups_claim = models.CharField(
        _("groups claim"),
        max_length=50,
        default="roles",
        help_text=_(
            "The name of the OIDC claim that holds the values to map to local user groups."
        ),
    )
    sync_groups = models.BooleanField(
        _("synchronize groups"),
        default=True,
        help_text=_(
            "Synchronize the local user groups with the provided groups. Note that this "
            "means a user is removed from all groups if there is no group claim. "
            "Uncheck to manage groups manually."
        ),
    )
    sync_groups_glob_pattern = models.CharField(
        _("groups glob pattern"),
        default="*",
        max_length=255,
        help_text=_(
            "The glob pattern that groups must match to be synchronized to "
            "the local database."
        ),
    )
    make_users_staff = models.BooleanField(
        _("make users staff"),
        default=False,
        help_text=_(
            "Users will be flagged as being a staff user automatically. This allows users to login to the admin interface. By default they have no permissions, even if they are staff."
        ),
    )

    class Meta:
        verbose_name = _("OpenID Connect configuration")

    def __str__(self):
        return force_text(self._meta.verbose_name)

    def clean(self):
        super().clean()

        # validate claim mapping
        User = get_user_model()
        for field in self.claim_mapping.keys():
            try:
                User._meta.get_field(field)
            except models.FieldDoesNotExist:
                raise ValidationError(
                    {
                        "claim_mapping": _(
                            "Field {field} does not exist on the user model"
                        ).format(field=field)
                    }
                )

        if User.USERNAME_FIELD in self.claim_mapping:
            raise ValidationError(
                {
                    "claim_mapping": _(
                        "The username field may not be in the claim mapping"
                    ),
                }
            )

    @property
    def oidc_rp_scopes(self):
        """
        Scopes should be formatted as a string with spaces
        """
        return " ".join(self.oidc_rp_scopes_list)

    @classmethod
    def clear_cache(cls):
        cache_name = getattr(
            settings, "OIDC_CACHE", oidc_settings.MOZILLA_DJANGO_OIDC_DB_CACHE
        )
        if cache_name:
            cache = get_cache(cache_name)
            cache_key = cls.get_cache_key()
            cache.delete(cache_key)

    def set_to_cache(self):
        cache_name = getattr(
            settings,
            "MOZILLA_DJANGO_OIDC_DB_CACHE",
            oidc_settings.MOZILLA_DJANGO_OIDC_DB_CACHE,
        )
        if not cache_name:
            return None
        cache = get_cache(cache_name)
        cache_key = self.get_cache_key()
        timeout = getattr(
            settings,
            "MOZILLA_DJANGO_OIDC_DB_CACHE_TIMEOUT",
            oidc_settings.MOZILLA_DJANGO_OIDC_DB_CACHE_TIMEOUT,
        )
        cache.set(cache_key, self, timeout)

    @classmethod
    def get_cache_key(cls):
        prefix = getattr(
            settings,
            "MOZILLA_DJANGO_OIDC_DB_PREFIX",
            oidc_settings.MOZILLA_DJANGO_OIDC_DB_PREFIX,
        )
        return "%s:%s" % (prefix, cls.__name__.lower())

    @classmethod
    def get_solo(cls):
        cache_name = getattr(
            settings,
            "MOZILLA_DJANGO_OIDC_DB_CACHE",
            oidc_settings.MOZILLA_DJANGO_OIDC_DB_CACHE,
        )
        if not cache_name:
            obj, created = cls.objects.get_or_create(pk=cls.singleton_instance_id)
            return obj
        cache = get_cache(cache_name)
        cache_key = cls.get_cache_key()
        obj = cache.get(cache_key)
        if not obj:
            obj, created = cls.objects.get_or_create(pk=cls.singleton_instance_id)
            obj.set_to_cache()
        return obj



# Generated by Django 2.2.24 on 2021-07-19 08:03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        (
            "mozilla_django_oidc_db",
            "0002_openidconnectconfig_oidc_op_discovery_endpoint",
        ),
    ]

    operations = [
        migrations.AlterField(
            model_name="openidconnectconfig",
            name="oidc_op_discovery_endpoint",
            field=models.URLField(
                blank=True,
                help_text="URL of your OpenID Connect provider discovery endpoint ending with a slash (`.well-known/...` will be added automatically). If this is provided, the remaining endpoints can be omitted, as they will be derived from this endpoint.",
                max_length=1000,
                verbose_name="Discovery endpoint",
            ),
        ),
    ]



from mozilla_django_oidc.views import (
    OIDCAuthenticationRequestView as _OIDCAuthenticationRequestView,
)

from .mixins import SoloConfigMixin


class OIDCAuthenticationRequestView(SoloConfigMixin, _OIDCAuthenticationRequestView):
    pass



from django.contrib import admin
from django.utils.translation import gettext_lazy as _

from django_better_admin_arrayfield.admin.mixins import DynamicArrayMixin
from solo.admin import SingletonModelAdmin

from .forms import OpenIDConnectConfigForm
from .models import OpenIDConnectConfig


@admin.register(OpenIDConnectConfig)
class OpenIDConnectConfigAdmin(DynamicArrayMixin, SingletonModelAdmin):
    form = OpenIDConnectConfigForm
    fieldsets = (
        (
            _("Activation"),
            {"fields": ("enabled",)},
        ),
        (
            _("Common settings"),
            {
                "fields": (
                    "oidc_rp_client_id",
                    "oidc_rp_client_secret",
                    "oidc_rp_scopes_list",
                    "oidc_rp_sign_algo",
                    "oidc_rp_idp_sign_key",
                )
            },
        ),
        (
            _("Endpoints"),
            {
                "fields": (
                    "oidc_op_discovery_endpoint",
                    "oidc_op_jwks_endpoint",
                    "oidc_op_authorization_endpoint",
                    "oidc_op_token_endpoint",
                    "oidc_op_user_endpoint",
                )
            },
        ),
        (
            _("User profile"),
            {
                "fields": (
                    "username_claim",
                    "groups_claim",
                    "claim_mapping",
                    "sync_groups",
                    "sync_groups_glob_pattern",
                    "make_users_staff",
                )
            },
        ),
    )



from mozilla_django_oidc.utils import import_from_settings

from .models import OpenIDConnectConfig


class SoloConfigMixin:
    @staticmethod
    def get_settings(attr, *args):
        config = OpenIDConnectConfig.get_solo()
        attr_lowercase = attr.lower()
        if getattr(config, attr_lowercase, ""):
            return getattr(config, attr_lowercase)
        return import_from_settings(attr, *args)



from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.test import RequestFactory

import pytest

from mozilla_django_oidc_db.backends import OIDCAuthenticationBackend
from mozilla_django_oidc_db.models import OpenIDConnectConfig


@pytest.mark.django_db
def test_backend_authenticate_oidc_not_enabled():
    config = OpenIDConnectConfig.get_solo()

    config.enabled = False
    config.save()

    backend = OIDCAuthenticationBackend()

    request = RequestFactory().get("/")

    # Authentication with the backend should not return a result,
    # because OIDC is not enabled
    assert backend.authenticate(request) is None


@pytest.mark.django_db
def test_backend_get_user_instance_values():
    claims = {
        "sub": "123456",
        "email": "admin@localhost",
        "given_name": "John",
        "family_name": "Doe",
    }

    backend = OIDCAuthenticationBackend()

    user_values = backend.get_user_instance_values(claims)

    assert user_values == {
        "email": "admin@localhost",
        "first_name": "John",
        "last_name": "Doe",
    }


@pytest.mark.django_db
def test_backend_create_user():
    config = OpenIDConnectConfig.get_solo()

    config.enabled = True
    config.oidc_rp_client_id = "testid"
    config.oidc_rp_client_secret = "secret"
    config.oidc_rp_sign_algo = "HS256"
    config.oidc_rp_scopes_list = ["openid", "email"]
    config.oidc_op_jwks_endpoint = "http://some.endpoint/v1/jwks"
    config.oidc_op_authorization_endpoint = "http://some.endpoint/v1/auth"
    config.oidc_op_token_endpoint = "http://some.endpoint/v1/token"
    config.oidc_op_user_endpoint = "http://some.endpoint/v1/user"
    config.save()

    User = get_user_model()

    claims = {
        "sub": "123456",
        "email": "admin@localhost",
        "given_name": "John",
        "family_name": "Doe",
    }

    backend = OIDCAuthenticationBackend()

    user = backend.create_user(claims)

    # Verify that a user is created with the correct values
    assert user.username == "123456"
    assert user.email == "admin@localhost"
    assert user.first_name == "John"
    assert user.last_name == "Doe"


@pytest.mark.django_db
def test_backend_create_user_different_username_claim():
    config = OpenIDConnectConfig.get_solo()

    config.enabled = True
    config.oidc_rp_client_id = "testid"
    config.oidc_rp_client_secret = "secret"
    config.oidc_rp_sign_algo = "HS256"
    config.oidc_rp_scopes_list = ["openid", "email"]
    config.oidc_op_jwks_endpoint = "http://some.endpoint/v1/jwks"
    config.oidc_op_authorization_endpoint = "http://some.endpoint/v1/auth"
    config.oidc_op_token_endpoint = "http://some.endpoint/v1/token"
    config.oidc_op_user_endpoint = "http://some.endpoint/v1/user"
    config.username_claim = "upn"
    config.save()

    User = get_user_model()

    claims = {
        "sub": "123456",
        "upn": "admin",
        "email": "admin@localhost",
        "given_name": "John",
        "family_name": "Doe",
    }

    backend = OIDCAuthenticationBackend()

    user = backend.create_user(claims)

    # Verify `upn` is used as username
    assert user.username == "admin"
    assert user.email == "admin@localhost"
    assert user.first_name == "John"
    assert user.last_name == "Doe"


@pytest.mark.django_db
def test_backend_filter_users():
    config = OpenIDConnectConfig.get_solo()

    config.enabled = True
    config.oidc_rp_client_id = "testid"
    config.oidc_rp_client_secret = "secret"
    config.oidc_rp_sign_algo = "HS256"
    config.oidc_rp_scopes_list = ["openid", "email"]
    config.oidc_op_jwks_endpoint = "http://some.endpoint/v1/jwks"
    config.oidc_op_authorization_endpoint = "http://some.endpoint/v1/auth"
    config.oidc_op_token_endpoint = "http://some.endpoint/v1/token"
    config.oidc_op_user_endpoint = "http://some.endpoint/v1/user"
    config.username_claim = "sub"
    config.save()

    User = get_user_model()

    # Create two users with the same email address, this shouldn't cause problems
    user1 = User.objects.create(
        username="123456", email="admin@localhost", first_name="John", last_name="Doe"
    )
    user2 = User.objects.create(
        username="654321", email="admin@localhost", first_name="Jane", last_name="Doe"
    )

    claims = {
        "sub": "123456",
        "email": "admin@localhost",
        "given_name": "John",
        "family_name": "Doe",
    }

    backend = OIDCAuthenticationBackend()

    users = backend.filter_users_by_claims(claims)

    assert users.count() == 1

    user = users.first()

    # Verify that a user is created with the correct values
    assert user.username == "123456"
    assert user.email == "admin@localhost"
    assert user.first_name == "John"
    assert user.last_name == "Doe"


@pytest.mark.django_db
def test_backend_filter_users_different_username_claim():
    config = OpenIDConnectConfig.get_solo()

    config.enabled = True
    config.oidc_rp_client_id = "testid"
    config.oidc_rp_client_secret = "secret"
    config.oidc_rp_sign_algo = "HS256"
    config.oidc_rp_scopes_list = ["openid", "email"]
    config.oidc_op_jwks_endpoint = "http://some.endpoint/v1/jwks"
    config.oidc_op_authorization_endpoint = "http://some.endpoint/v1/auth"
    config.oidc_op_token_endpoint = "http://some.endpoint/v1/token"
    config.oidc_op_user_endpoint = "http://some.endpoint/v1/user"
    config.username_claim = "upn"
    config.save()

    User = get_user_model()

    # Create two users with the same email address, this shouldn't cause problems
    user1 = User.objects.create(
        username="admin", email="admin@localhost", first_name="John", last_name="Doe"
    )
    user2 = User.objects.create(
        username="admin2", email="admin@localhost", first_name="Jane", last_name="Doe"
    )

    claims = {
        "upn": "admin2",
        "email": "admin@localhost",
        "given_name": "Jane",
        "family_name": "Doe",
    }

    backend = OIDCAuthenticationBackend()

    users = backend.filter_users_by_claims(claims)

    assert users.count() == 1

    user = users.first()

    # Verify that a user is created with the correct values
    assert user.username == "admin2"
    assert user.email == "admin@localhost"
    assert user.first_name == "Jane"
    assert user.last_name == "Doe"


@pytest.mark.django_db
def test_backend_update_user():
    config = OpenIDConnectConfig.get_solo()

    config.enabled = True
    config.oidc_rp_client_id = "testid"
    config.oidc_rp_client_secret = "secret"
    config.oidc_rp_sign_algo = "HS256"
    config.oidc_rp_scopes_list = ["openid", "email"]
    config.oidc_op_jwks_endpoint = "http://some.endpoint/v1/jwks"
    config.oidc_op_authorization_endpoint = "http://some.endpoint/v1/auth"
    config.oidc_op_token_endpoint = "http://some.endpoint/v1/token"
    config.oidc_op_user_endpoint = "http://some.endpoint/v1/user"
    config.username_claim = "sub"
    config.save()

    User = get_user_model()

    # Create two users with the same email address, this shouldn't cause problems
    user1 = User.objects.create(
        username="123456", email="admin@localhost", first_name="John", last_name="Doe"
    )
    user2 = User.objects.create(
        username="654321", email="admin@localhost", first_name="Jane", last_name="Doe"
    )

    claims = {
        "sub": "123456",
        "email": "modified@localhost",
        "given_name": "Name",
        "family_name": "Modified",
    }

    backend = OIDCAuthenticationBackend()

    user = backend.update_user(user1, claims)

    # Verify that a user is created with the correct values
    assert user.username == "123456"
    assert user.email == "modified@localhost"
    assert user.first_name == "Name"
    assert user.last_name == "Modified"


@pytest.mark.django_db
def test_backend_create_user_sync_all_groups():
    config = OpenIDConnectConfig.get_solo()

    config.enabled = True
    config.groups_claim = "roles"
    config.sync_groups = True
    config.sync_groups_glob_pattern = "*"
    config.save()

    claims = {
        "sub": "123456",
        "roles": ["useradmin", "groupadmin"],
    }

    backend = OIDCAuthenticationBackend()

    user = backend.create_user(claims)

    # Verify that the groups were created
    assert Group.objects.count() == 2

    # Verify that a user is created with the correct values
    assert user.username == "123456"
    assert list(user.groups.values_list("name", flat=True)) == [
        "useradmin",
        "groupadmin",
    ]


@pytest.mark.django_db
def test_backend_create_user_sync_groups_according_to_pattern():
    Group.objects.all().delete()
    config = OpenIDConnectConfig.get_solo()

    config.enabled = True
    config.groups_claim = "roles"
    config.sync_groups = True
    config.sync_groups_glob_pattern = "group*"
    config.save()

    claims = {
        "sub": "123456",
        "roles": ["useradmin", "groupadmin"],
    }

    backend = OIDCAuthenticationBackend()

    user = backend.create_user(claims)

    # Verify that a user is created with the correct values
    assert user.username == "123456"
    assert list(user.groups.values_list("name", flat=True)) == ["groupadmin"]


@pytest.mark.django_db
def test_backend_create_user_with_profile_settings():
    Group.objects.all().delete()
    config = OpenIDConnectConfig.get_solo()

    config.enabled = True
    config.groups_claim = "roles"
    config.sync_groups = True
    config.claim_mapping = {
        "first_name": "given_name",
        "last_name": "family_name",
        "email": "email",
        "is_superuser": "is_god",
    }
    config.sync_groups_glob_pattern = "*"
    config.make_users_staff = True
    config.save()

    Group.objects.create(name="useradmin")
    Group.objects.create(name="groupadmin")

    claims = {
        "sub": "123456",
        "email": "admin@localhost",
        "given_name": "John",
        "family_name": "Doe",
        "is_god": 1,
        "roles": [
            "useradmin",
        ],
    }

    backend = OIDCAuthenticationBackend()

    user = backend.create_user(claims)

    # Verify that a user is created with the correct values
    assert user.username == "123456"
    assert user.email == "admin@localhost"
    assert user.first_name == "John"
    assert user.last_name == "Doe"
    assert user.is_staff == True
    assert user.is_superuser == True
    assert list(user.groups.values_list("name", flat=True)) == ["useradmin"]
